// **********************************************************************
//          Java application to extract protein surface residues
//
//   This application utilizes an external program, 'Surface Racer' to
//   calculate solvent accessible surface area.
//
//            developed by Hui Sun Lee (huisun.cadd@gmail.com)
// **********************************************************************

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;



public class GetSurfaceResiduesUsingSurfaceRacer{
	public static void main( String[] args ) throws IOException{
		if( args.length == 0 || args.length != 3 || args[0].equalsIgnoreCase( "--help" ) ){
			System.out.println( "Usage: GetSurfaceResiduesUsingSurfaceRacer" );
			System.out.println( "            -opts  [protein structure (PDB)]" );
			System.out.println( "                   [relative solvent accessible surface area (rSASA) (%)]" );
			System.out.println( "                   [SurfaceRacer execution file name]" );

			return;
		}

		
		
		
		String mProtein = args[0];
		double rSASA = Double.parseDouble( args[1] );   
		String mSurfaceRacer_Exe = args[2];  


		
		FileInputStream list_stream;
		BufferedReader list_reader;	
		String list_line;
		
		FileInputStream struct_stream;
		BufferedReader struct_reader;	
		String struct_line;
		
		DataOutputStream out;	
		

		Runtime rt = Runtime.getRuntime();
		Process pr;
		LineNumberReader output_log;
		String pr_line;
		String command;
		

		Hashtable<String,String> htResIndex = null;
		Vector<String> vResIndex = null;
		boolean [] isSelectedResidue = new boolean[2000];
		double [] sasaResidue = new double[2000];

		String [] array;		
		String value = null;



		try{
			// =================================================================================================================
			//  A residue is counted as solvent-accessible if the relative solvent accessible solvent area (rSASA) of a residue 
			//  (SASA divided by its maximum SASA) is at least X% of its maximum area.
			// =================================================================================================================

			System.out.println( "Extracting surface residues for " + mProtein );
				
						
			String protein = mProtein;
			String surface = mProtein.substring( 0, mProtein.length()-4 ) + "-surface.pdb";
			
				              
            // ****************************************************************************************************************************
            // Surface Racer web site: http://pharmacy.mc.uky.edu/faculty/tsodikov/software.php
            // Reference:
            // Tsodikov, O. V., Record, M. T. Jr. and Sergeev, Y. V. (2002). 
            // A novel computer program for fast exact calculation of accessible and molecular surface areas and average surface curvature. 
            // J. Comput. Chem., 23, 600-609 
            // ****************************************************************************************************************************
                
            String surfrace_sasa_res = mProtein.substring( 0, mProtein.length()-4 ) + "_residue.txt";
			String surfrace_sasa_atm = mProtein.substring( 0, mProtein.length()-4 ) + ".txt";
				
				
			out = new DataOutputStream( new FileOutputStream( new File( "surface_racer.in" ) ) );	
			out.writeBytes( "1\n" );
			out.writeBytes( protein + "\n" );	
			out.writeBytes( "1.4\n" );
			out.writeBytes( "3\n" );
			out.writeBytes( "\n" );
			out.writeBytes( "\n" );
			out.close();
				
            pr = rt.exec( new String[] { "/bin/sh", "-c", mSurfaceRacer_Exe + " < surface_racer.in" } );
			pr.waitFor();
			pr.getInputStream().close();
			pr.destroy();  
               
                
                
                
                
            // *******************************
            //    identify surface residues
            // *******************************
                
            htResIndex = new Hashtable<String,String>();
            vResIndex = new Vector<String>();
                
            for( int i = 0 ; i < isSelectedResidue.length ; i++ ){
              	isSelectedResidue[i] = false;
               	sasaResidue[i] = 0;
            }
                
                
                
                
            // *** set resSeq index ***
            struct_stream = new FileInputStream( new File( protein ) );
			struct_reader = new BufferedReader( new InputStreamReader( struct_stream ) );	
				
			String resName;
			String resSeq;
			int index = -1;

			while( ( struct_line = struct_reader.readLine()) != null ){	
				if( struct_line.startsWith( "TER" ) )
					break;
						
				resName = struct_line.substring( 17, 20 );
				resSeq = struct_line.substring( 22, 27 ).trim();
				value = htResIndex.get( resSeq );
					
				if( value == null ){
    				index++;
    				htResIndex.put( resSeq, "" + index );
    				vResIndex.add( resSeq + " " + resName );
    			}
			}
				
			struct_reader.close();
			struct_stream.close();
				
				
				
				
			// *** calculate SASA per residue ***
            struct_stream = new FileInputStream( new File( surfrace_sasa_atm ) );
			struct_reader = new BufferedReader( new InputStreamReader( struct_stream ) );	
				
			while( ( struct_line = struct_reader.readLine()) != null ){	
				if( struct_line.startsWith( "CAVITY" ) )
					break;
    				
    			resSeq = struct_line.substring( 22, 27 ).trim();
    			value = htResIndex.get( resSeq );
    			sasaResidue[Integer.parseInt(value)] += Double.parseDouble( struct_line.substring( 62, 68 ).trim() );
			}
				
			struct_reader.close();
			struct_stream.close();
				
				
				
				
			// *** identify surface residues based on rSASA ***
			for( int i = 0 ; i < vResIndex.size() ; i++ ){
				array = vResIndex.elementAt(i).split( " " );
					
				value = htResIndex.get( array[0] );					
				double max_sasa = getMaxSASA( array[1] );
					
				if( ( sasaResidue[Integer.parseInt(value)]/max_sasa * 100 ) >= rSASA ){
					isSelectedResidue[Integer.parseInt(value)] = true;
				}
			}
				
				
				
			/*
			// *** identify surface residues based on cavities ***
			struct_stream = new FileInputStream( new File( surfrace_sasa_atm ) );
			struct_reader = new BufferedReader( new InputStreamReader( struct_stream ) );	
				
			while( ( struct_line = struct_reader.readLine()) != null ){	
				if( struct_line.startsWith( "CAVITY" ) ){
					while( ( struct_line = struct_reader.readLine()) != null ){	
						if( struct_line.startsWith( "ATOM" ) ){
							resSeq = struct_line.substring( 22, 27 ).trim();
    						value = htResIndex.get( resSeq );
    						isSelectedResidue[Integer.parseInt(value)] = true;
    					}
					}
				}
			}
				
			struct_reader.close();
			struct_stream.close();
			*/
				
				
				
			// *** write surface residues ***
			out = new DataOutputStream( new FileOutputStream( new File( surface ) ) );	
				
			struct_stream = new FileInputStream( new File( protein ) );
			struct_reader = new BufferedReader( new InputStreamReader( struct_stream ) );	
				
			while( ( struct_line = struct_reader.readLine()) != null ){	
				if( struct_line.startsWith( "TER" ) )
					break;
						
				resSeq = struct_line.substring( 22, 27 ).trim();
				value = htResIndex.get( resSeq );
					
				if( isSelectedResidue[Integer.parseInt(value)] )
					out.writeBytes( struct_line + "\n" );
			}
				
			out.writeBytes( "TER" + "\n" );
				
			struct_reader.close();
			struct_stream.close();              
               
            out.close();
		}
		catch( Exception exception ){
			exception.printStackTrace();
		}
	}	
	



	public static double getMaxSASA( String res_name ){
	    // **************************************************************************************************************************************
		//    setting maximumal solvent accessible surface area (A^2) for each residue
		//
		//    Reference:
		//    C. M. Topham and J. C. Smith
		//    Tri-peptide reference structures for the calculation of relative solvent accessible surface area in protein amino acid residues
		//    Comput. Biol. Chem. 54 (2015) 33-43
		// *************************************************************************************************************************************
		
		if( res_name.equalsIgnoreCase( "ALA" ) )
			return 108.50;
		else if( res_name.equalsIgnoreCase( "ARG" ) )
			return 246.37;
		else if( res_name.equalsIgnoreCase( "ASN" ) )
			return 156.64;
		else if( res_name.equalsIgnoreCase( "ASP" ) )
			return 146.17;
		else if( res_name.equalsIgnoreCase( "CYS" ) )
			return 138.40;
		else if( res_name.equalsIgnoreCase( "GLN" ) )
			return 181.94;
		else if( res_name.equalsIgnoreCase( "GLU" ) )
			return 177.12;
		else if( res_name.equalsIgnoreCase( "GLY" ) )
			return 81.30;
		else if( res_name.equalsIgnoreCase( "HIS" ) )
			return 184.51;
		else if( res_name.equalsIgnoreCase( "ILE" ) )
			return 181.50;
		else if( res_name.equalsIgnoreCase( "LEU" ) )
			return 182.10;
		else if( res_name.equalsIgnoreCase( "LYS" ) )
			return 204.85;
		else if( res_name.equalsIgnoreCase( "MET" ) )
			return 199.13;
		else if( res_name.equalsIgnoreCase( "PHE" ) )
			return 208.47;
		else if( res_name.equalsIgnoreCase( "PRO" ) )  // average of Pro_down and Pro_up
			return 139.44;
		else if( res_name.equalsIgnoreCase( "SER" ) )  // average of Ser_trans and Ser_g-
			return 117.71;
		else if( res_name.equalsIgnoreCase( "THR" ) )
			return 140.10;
		else if( res_name.equalsIgnoreCase( "TRP" ) )
			return 253.79;	
		else if( res_name.equalsIgnoreCase( "TYR" ) )
			return 216.69;
		else if( res_name.equalsIgnoreCase( "VAL" ) )
			return 152.69;
		else
			return 170.87;  // average over 20 standard amino acids
	}
}
	